@@ -9,7 +9,7 @@ from torch import nn, einsum
 import torch.nn.functional as F
 from torch.fft import fft2, ifft2
 
-from einops import rearrange, reduce
+from einops import rearrange, reduce, pack
 from einops.layers.torch import Rearrange
 
 from tqdm.auto import tqdm
@@ -198,16 +198,29 @@ class Attention(nn.Module):
 
 # conditioning class
 
-class FourierConditioning(nn.Module):
-    def __init__(self, dim):
+class Conditioning(nn.Module):
+    def __init__(self, fmap_size, dim):
         super().__init__()
+        self.ff_parser_attn_map = nn.Parameter(torch.ones(dim, fmap_size, fmap_size))
+
         self.norm_input = LayerNorm(dim, bias = True)
         self.norm_condition = LayerNorm(dim, bias = True)
 
     def forward(self, x, c):
+
+        # ff-parser in the paper, for modulating out the high frequencies
+
+        dtype = x.dtype
+        x = fft2(x)
+        x = x * self.ff_parser_attn_map
+        x = ifft2(x).real
+        x = x.type(dtype)
+
+        # eq 3 in paper
+
         normed_x = self.norm_input(x)
         normed_c = self.norm_condition(c)
-        return (normed_x * normed_c) * c   # eq 3 in paper
+        return (normed_x * normed_c) * c
 
 # model
 
@@ -215,16 +228,20 @@ class Unet(nn.Module):
     def __init__(
         self,
         dim,
+        image_size,
         init_dim = None,
         out_dim = None,
         dim_mults=(1, 2, 4, 8),
         channels = 3,
         self_condition = False,
         resnet_block_groups = 8,
-        conditioning_klass = FourierConditioning
+        conditioning_klass = Conditioning,
+        skip_connect_condition_fmaps = False   # whether to concatenate the conditioning fmaps in the latter decoder upsampling portion of unet
     ):
         super().__init__()
 
+        self.image_size = image_size
+
         # determine dimensions
 
         self.channels = channels
@@ -258,14 +275,18 @@ class Unet(nn.Module):
 
         self.conditioners = nn.ModuleList([])
 
+        self.skip_connect_condition_fmaps = skip_connect_condition_fmaps
+
         # downsampling encoding blocks
 
         self.downs = nn.ModuleList([])
 
+        curr_fmap_size = image_size
+
         for ind, (dim_in, dim_out) in enumerate(in_out):
             is_last = ind >= (num_resolutions - 1)
 
-            self.conditioners.append(conditioning_klass(dim_in))
+            self.conditioners.append(conditioning_klass(curr_fmap_size, dim_in))
 
             self.downs.append(nn.ModuleList([
                 block_klass(dim_in, dim_in, time_emb_dim = time_dim),
@@ -274,6 +295,9 @@ class Unet(nn.Module):
                 Downsample(dim_in, dim_out) if not is_last else nn.Conv2d(dim_in, dim_out, 3, padding = 1)
             ]))
 
+            if not is_last:
+                curr_fmap_size //= 2
+
         # middle blocks
 
         mid_dim = dims[-1]
@@ -293,9 +317,11 @@ class Unet(nn.Module):
         for ind, (dim_in, dim_out) in enumerate(reversed(in_out)):
             is_last = ind == (len(in_out) - 1)
 
+            skip_connect_dim = dim_in * (2 if self.skip_connect_condition_fmaps else 1)
+
             self.ups.append(nn.ModuleList([
-                block_klass(dim_out + dim_in, dim_out, time_emb_dim = time_dim),
-                block_klass(dim_out + dim_in, dim_out, time_emb_dim = time_dim),
+                block_klass(dim_out + skip_connect_dim, dim_out, time_emb_dim = time_dim),
+                block_klass(dim_out + skip_connect_dim, dim_out, time_emb_dim = time_dim),
                 Residual(LinearAttention(dim_out)),
                 Upsample(dim_out, dim_in) if not is_last else  nn.Conv2d(dim_out, dim_in, 3, padding = 1)
             ]))
@@ -312,7 +338,7 @@ class Unet(nn.Module):
         cond,
         x_self_cond = None
     ):
-        dtype = x.dtype
+        dtype, skip_connect_c = x.dtype, self.skip_connect_condition_fmaps
 
         if self.self_condition:
             x_self_cond = default(x_self_cond, lambda: torch.zeros_like(x))
@@ -331,7 +357,7 @@ class Unet(nn.Module):
             x = block1(x, t)
             c = cond_block1(c, t)
 
-            h.append(x)
+            h.append([x, c] if skip_connect_c else [x])
 
             x = block2(x, t)
             c = cond_block2(c, t)
@@ -344,7 +370,7 @@ class Unet(nn.Module):
 
             c = conditioner(x, c)
 
-            h.append(x)
+            h.append([x, c] if skip_connect_c else [x])
 
             x = downsample(x)
             c = cond_downsample(c)
@@ -358,10 +384,10 @@ class Unet(nn.Module):
         x = self.mid_block2(x, t)
 
         for block1, block2, attn, upsample in self.ups:
-            x = torch.cat((x, h.pop()), dim = 1)
+            x = torch.cat((x, *h.pop()), dim = 1)
             x = block1(x, t)
 
-            x = torch.cat((x, h.pop()), dim = 1)
+            x = torch.cat((x, *h.pop()), dim = 1)
             x = block2(x, t)
             x = attn(x)
 
@@ -402,7 +428,6 @@ class MedSegDiff(nn.Module):
         self,
         model,
         *,
-        image_size,
         timesteps = 1000,
         sampling_timesteps = None,
         objective = 'pred_noise',
@@ -415,7 +440,7 @@ class MedSegDiff(nn.Module):
         self.channels = self.model.channels
         self.self_condition = self.model.self_condition
 
-        self.image_size = image_size
+        self.image_size = model.image_size
 
         self.objective = objective
 
